#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging

from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.interfaces import ExecutionManagerIF

logger = logging.getLogger(__name__)


class CapabilityQueueManager:
    """
    This is a prototype of a new kind of capability queue manager that does not hold onto any
    internal state itself. All of its calculations are performed against a certain database
    query, which is re-run whenever something "interesting" happens. The API of the class
    is basically to provide a way to send signals when "interesting" things happen, so it
    can react by checking to see if we're ready to execute something or not.
    """

    def __init__(self, queue_reporter: CapabilityInfo, executor: ExecutionManagerIF):
        self.queue_reporter = queue_reporter
        self.executor = executor

    def execution_requested(self):
        self.launch_if_availability()

    def execution_complete(self):
        self.launch_if_availability()

    def launch_if_availability(self):
        logger.info("Obtaining queue report")
        for queue in self.queue_reporter.report():
            logger.info(
                "Considering capability queue %s with %s allowed, %s executing and %s waiting",
                queue.capability,
                queue.concurrency_allowed,
                queue.executing,
                queue.waiting,
            )
            # how many jobs do we start? if we have an arbitrarily large number, we can have
            # as many as there are allowed by the concurrency limit. But some are probably
            # already executing, so we deduct the executing number.
            #
            # What if we don't have enough jobs to fulfill the demand? That's OK, we won't
            # launch more jobs than we have available.
            #
            # Thus, the minimum of what is still allowed and what is available
            jobs_to_run = min(queue.concurrency_allowed - queue.executing, queue.waiting)
            if jobs_to_run > 0:
                for execution in self.queue_reporter.waiting_executions(queue.capability, jobs_to_run):
                    logger.info("Waking up execution %s", execution.id)
                    self.executor.resume_execution(execution)

    def reconfigure(self):
        self.startup()

    def startup(self):
        self.launch_if_availability()
